// Mega 644 version

//Run all three timers in different modes
// timer 0 -- compare-match timebase
// timer 1 -- input capture from analog comparator
// timer 2 -- stand alone square wave gen  

// main will schedule task1, and poll the ACSR ACO bit
// task 1 will  print the periods 
// timer0 compare ISR will increment a timer
// timer1 capture ISR will compute a period from the capture

// MUST connect port B.3 (AIN1) to port D.7 (OC2)
// Capture yields cycle-accurate period measurement
// Polling the ACO bit yields 20 cycle variability 
                
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>

//set up the debugging utility ASSERT
//#define __ASSERT_USE_STDERR
//#include <assert.h>

#include "uart.h"
#include "uart.c"

#include "clock.h"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
       

//time for task
#define t1 250
#define t2 250
#define t3 300
#define t4 200
#define t5 50  

#define begin {
#define end }

#define sev0 0x7e
#define sev1 0x30
#define sev2 0x6d
#define sev3 0x79
#define sev4 0x33
#define sev5 0x5b
#define sev6 0x5f
#define sev7 0x70
#define sev8 0x7f
#define sev9 0x7b

#define PORTLED PORTA
#define PORTSTK PORTC
#define SETALARM PIND2

//the task subroutine
void task1(void);
void task2(void);
void task3(void);
void task4(void);
void task5(void);
void task6(void);

void initialize(void); //all the usual mcu stuff 

//timeout counter for task1 and task2        
volatile unsigned char time1; 
volatile unsigned char time2;
volatile unsigned char time3;
volatile unsigned char time4;
volatile unsigned char time5;

unsigned int reading; //accelerometer reading

char AinH, AinL; //high and low registers for ADC

int milliCounter;
int secondsCounter;
int minutesCounter;
int hoursCounter;
unsigned char ampmFlag, a_ampmFlag;

unsigned char alarmFlag;
unsigned char alarmContinueFlag;
unsigned char snoozeFlag;


unsigned char cminL, chourL; //represents the lower digit of the minutes
unsigned char cminH, chourH; //represents the higher digit of the minutes

unsigned char aminL, aminH, ahourL, ahourH;
unsigned char amin, ahour;

unsigned char clock; //clock flag, 1: trying to set the clock, 0: trying to set the alarm

unsigned char minutesButtonFlag; //flag for checking to see if a button has been pressed
unsigned char hoursButtonFlag;
unsigned char digitButtonFlag;
unsigned char alarmButtonFlag; 
unsigned char snoozeButtonFlag;


/* States for Button Push State Machines */

//State machine state names
#define NoPush 1 
#define MaybePush 2
#define Pushed 3
#define MaybeNoPush 4

unsigned char minutesButtonPushState;
unsigned char hoursButtonPushState;
unsigned char digitButtonPushState;
unsigned char alarmButtonPushState;
unsigned char snoozeButtonPushState;

unsigned char digit; // digit flag, 0: minL, 1: minH, 2: hourL, 3: hourH and TURNS on corresponding STK500 LED
unsigned char led;

/* End States */

//**********************************************************
//timer 0 overflow ISR
ISR (TIMER0_COMPA_vect) 
begin      
  //Decrement the time if not already zero
  if (time1>0)	--time1;
  if (time2>0)  --time2;
  if (time3>0)  --time3;
  if (time4>0)  --time4;
  if (time5>0)  --time5;

  milliCounter++;

  if(milliCounter == 1000) {
  	secondsCounter++;
	milliCounter = 0;
	if(secondsCounter == 60) {
		minutesCounter++;
		secondsCounter = 0;
		if(minutesCounter == 60) {
			hoursCounter ++;
			minutesCounter = 0;
			if(hoursCounter == 13) {
				hoursCounter = 1;
			}
			if(hoursCounter == 12) {
				if(ampmFlag) { ampmFlag = 0; }
				else ampmFlag = 1;
			}
		}
	}
  }

end  


//**********************************************************       
//Entry point and task scheduler loop
int main(void)
begin  
  initialize();
  
  // main task scheduler loop 
  while(1)
  begin
  	// task1 prints the period     
    if (time1==0){time1=t1; task1();}
	if (time2==0){time2=t2; task2();}
	if (time3==0){time3=t3; task3();}
	if (time4==0){time4=t4; task4();}
	if (time5==0){time5=t5; task5();}
  end
end  
  
//**********************************************************          
//Timed task subroutine
//Task 1
//Print accelerometer reading to hyperterm
void task1(void) 
begin  
    //reset the task timer
	unsigned char tempmin, temphour;
  
  if (clock) {
  		if ((alarmFlag && (chourH==ahourH) && (chourL==ahourL) && (cminH==aminH) && (cminL==aminL)) || (alarmFlag && alarmContinueFlag)) {
			fprintf(stdout,"%d%d:%d%d:%d (%d) AL%d&%ds%d  %d  d%d ***** ALARM SOUNDING! ALARM SOUNDING! *****\n\r", 
									chourH, chourL,
  									cminH, cminL,
									secondsCounter, ampmFlag, alarmFlag, alarmContinueFlag, snoozeFlag,
									reading, digit);
			if (alarmFlag && (secondsCounter==59)) alarmContinueFlag = 1;
			if (alarmFlag && snoozeFlag) {
				tempmin = amin;
				temphour = ahour;
				if (amin <= 51) amin = amin + 8;
				else { 
					amin = (amin + 8) % 60;
					ahour++;
					if (ahour == 13) {
						ahour = 1;
					}
					if (ahour == 12) {
						if (a_ampmFlag) { a_ampmFlag = 0; }
						else a_ampmFlag = 1;
					}	
				}
				//snoozeFlag = 0;
				alarmContinueFlag = 0;
			}
			//if (!alarmFlag) alarmContinueFlag = 0;
		} else {
			if (!alarmFlag) alarmContinueFlag = 0;
	  		// print capture interval 
	  		fprintf(stdout,"%d%d:%d%d:%d (%d) AL%d&%ds%d  %d  d%d\n\r", 
										chourH, chourL,
	  									cminH, cminL,
										secondsCounter, ampmFlag, alarmFlag, alarmContinueFlag, snoozeFlag,
										reading, digit);
		}
  } else {
  		fprintf(stdout, "AL> %d%d:%d%d (%d) AL%d  d%d\n\r", ahourH, ahourL, aminH, aminL, a_ampmFlag, alarmFlag, digit);
  }
end



//Task 2
//Take A-to-D measurement
void task2(void)
begin
	//read in low before reading in high
	AinL = ADCL;
	AinH = ADCH;
	reading = 0;
	reading = AinH << 8;
	reading += AinL;

 	//start another conversion
 	ADCSRA |= (1<<ADSC) ; 
end  

//task 3 updates LED display for numbers
//runs every 300 ms
void task3(void)
begin

	unsigned char dispml, dispmh, disphl, disphh;
	unsigned char minL, minH, hourL, hourH;

	if(clock) {
		minL = cminL = minutesCounter % 10;
		minH = cminH = minutesCounter / 10;
		hourL = chourL = hoursCounter % 10;
		hourH = chourH = hoursCounter / 10;
	} else {
		minL = aminL = amin % 10;
		minH = aminH = amin / 10;
		hourL = ahourL = ahour % 10;
		hourH = ahourH = ahour / 10;
	}
	
	switch (minL) {
			case 1:
				dispml = sev1;
				break;
			case 2:
				dispml = sev2;
				break;
			case 3:
				dispml = sev3;
				break;
			case 4:
				dispml = sev4;
				break;
			case 5:
				dispml = sev5;
				break;
			case 6:
				dispml = sev6;
				break;
			case 7:
				dispml = sev7;
				break;
			case 8:
				dispml = sev8;
				break;
			case 9:
				dispml = sev9;
				break;
			case 0:
				dispml = sev0;
				break;
			default:
                dispml = sev8;
                break;
	}
	switch (minH) {
           case 1:
                dispmh = sev1;
                break;
           case 2:
                dispmh = sev2;
                break;
           case 3:
                dispmh = sev3;
                break;
           case 4:
                dispmh = sev4;
                break;
           case 5:
                dispmh = sev5;
                break;
		   case 0:
		   		dispmh = sev0;
				break;
           default:
                dispmh = sev8;
                break;
    }
    switch (hourL) {
          	case 1:
				disphl = sev1;
				break;
			case 2:
				disphl = sev2;
				break;
			case 3:
				disphl = sev3;
				break;
			case 4:
				disphl = sev4;
				break;
			case 5:
				disphl = sev5;
				break;
			case 6:
				disphl = sev6;
				break;
			case 7:
				disphl = sev7;
				break;
			case 8:
				disphl = sev8;
				break;
			case 9:
				disphl = sev9;
				break;
			case 0:
				disphl = sev0;
				break;
			default:
                disphl = sev8;
                break;
    }
    switch (hourH) {
           case 1:
                disphh = sev1;
                break;
           case 0:
                disphh = 0x00;
                break;
           default:
                disphh = sev8;
                break;
    }

	if (digit==3) { 
		PORTLED = disphh; led = 0xf7;
	} else if (digit==2) { 
         PORTLED = disphl; led = 0xfb;
	} else if (digit==1) { 
         PORTLED = dispmh; led = 0xfd;
	} else { 
         PORTLED = dispml; led = 0xfe; }
	
	if (alarmFlag) led = led ^ 0x80;
	if ((clock && ampmFlag) || (!clock && a_ampmFlag)) led = led ^ 0x20;
         
    PORTSTK = led;
end


//takes care of button flags
void task4(void)
begin
	if(minutesButtonFlag) {
		if(clock) {
			minutesCounter++;
			if(minutesCounter == 60) {
				minutesCounter = 0;
			}
			//minutesButtonFlag = 0;
		}
		else {
			amin++;
			if (amin == 60) {
				amin = 0;
			}
			//minutesButtonFlag = 0;
		}
		minutesButtonFlag = 0;
	}

	if(hoursButtonFlag) {
		fprintf(stdout, "Hours button pressed\n\r");
		if(clock) {
			hoursCounter++;
			if(hoursCounter == 13) {
				hoursCounter = 1;
			
			}
			if(hoursCounter == 12) {
				if(ampmFlag) { ampmFlag = 0; }
				else ampmFlag = 1;
			}
			//hoursButtonFlag = 0;
		}
		else {
			ahour++;
			if (ahour == 13) {
				ahour = 1;
			}
			if (ahour == 12) {
				if (a_ampmFlag) { a_ampmFlag = 0; }
				else a_ampmFlag = 1;
			}
			//hoursButtonFlag = 0;
		}
		hoursButtonFlag = 0;
	}

	if (digitButtonFlag) {
		if (digit==3) digit = 0;
		else digit++;
	}

	if (alarmButtonFlag) {
		alarmFlag = !alarmFlag;
	}

	if (snoozeButtonFlag && alarmFlag) {
		snoozeFlag = !snoozeFlag;
	}


end

void task5(void)
begin
  switch (minutesButtonPushState)
  begin
     case NoPush: 
        if (~PIND & 0x04) minutesButtonPushState=MaybePush;
        else minutesButtonPushState=NoPush;
        break;
     case MaybePush:
        if (~PIND & 0x04) 
        begin
           minutesButtonPushState=Pushed;   
           minutesButtonFlag=1;
        end
        else minutesButtonPushState=NoPush;
        break;
     case Pushed:
	 	if (~PIND & 0x04) minutesButtonPushState=Pushed; 
        else minutesButtonPushState=MaybeNoPush;    
        break;
     case MaybeNoPush:
        if (~PIND & 0x04) minutesButtonPushState=Pushed; 
        else 
        begin
           minutesButtonPushState=NoPush;
           minutesButtonFlag=0;
        end    
        break;
  end


  switch (hoursButtonPushState)
  begin
     case NoPush: 
        if (~PIND & 0x08) hoursButtonPushState=MaybePush;
        else hoursButtonPushState=NoPush;
        break;
     case MaybePush:
        if (~PIND & 0x08) 
        begin
           hoursButtonPushState=Pushed;   
           hoursButtonFlag=1;
        end
        else hoursButtonPushState=NoPush;
        break;
     case Pushed:
	 	if (~PIND & 0x08) hoursButtonPushState=Pushed; 
        else hoursButtonPushState=MaybeNoPush;    
        break;
     case MaybeNoPush:
        if (~PIND & 0x08) hoursButtonPushState=Pushed; 
        else 
        begin
           hoursButtonPushState=NoPush;
           hoursButtonFlag=0;
        end    
        break;
  end
  
  
  switch (digitButtonPushState)
  begin
     case NoPush: 
        if (~PIND & 0x10) digitButtonPushState=MaybePush;
        else digitButtonPushState=NoPush;
        break;
     case MaybePush:
        if (~PIND & 0x10) 
        begin
           digitButtonPushState=Pushed;   
           digitButtonFlag=1;
        end
        else digitButtonPushState=NoPush;
        break;
     case Pushed:
	 	if (~PIND & 0x10) digitButtonPushState=Pushed; 
        else digitButtonPushState=MaybeNoPush;    
        break;
     case MaybeNoPush:
        if (~PIND & 0x10) digitButtonPushState=Pushed; 
        else 
        begin
           digitButtonPushState=NoPush;
           digitButtonFlag=0;
        end    
        break;
  end


  switch (alarmButtonPushState)
  begin
     case NoPush: 
        if (~PIND & 0x20) alarmButtonPushState=MaybePush;
        else alarmButtonPushState=NoPush;
        break;
     case MaybePush:
        if (~PIND & 0x20) 
        begin
           alarmButtonPushState=Pushed;   
           alarmButtonFlag=1;
        end
        else alarmButtonPushState=NoPush;
        break;
     case Pushed:
	 	if (~PIND & 0x20) alarmButtonPushState=Pushed; 
        else alarmButtonPushState=MaybeNoPush;    
        break;
     case MaybeNoPush:
        if (~PIND & 0x20) alarmButtonPushState=Pushed; 
        else 
        begin
           alarmButtonPushState=NoPush;
           alarmButtonFlag=0;
        end    
        break;
  end

  switch (snoozeButtonPushState)
  begin
     case NoPush: 
        if (~PIND & 0x80) snoozeButtonPushState=MaybePush;
        else snoozeButtonPushState=NoPush;
        break;
     case MaybePush:
        if (~PIND & 0x80) 
        begin
           snoozeButtonPushState=Pushed;   
           snoozeButtonFlag=1;
        end
        else snoozeButtonPushState=NoPush;
        break;
     case Pushed:
	 	if (~PIND & 0x80) snoozeButtonPushState=Pushed; 
        else snoozeButtonPushState=MaybeNoPush;    
        break;
     case MaybeNoPush:
        if (~PIND & 0x80) snoozeButtonPushState=Pushed; 
        else 
        begin
           snoozeButtonPushState=NoPush;
           snoozeButtonFlag=0;
        end    
        break;
  end


  clock = ~PIND & 0x40;

end

//********************************************************** 
//Set it all up
void initialize(void)
begin
           
  //set up the ports
  DDRD = 0x00;	// PORT D is an input
  
  //PORTD = 0x0c; // 0b00000100
  PORTD = 0b11111100;  

  //DDRB = 0x00;
  //PORTB = 0b00000001;

  DDRC = 0xff;    // PORT C is an ouput 
  //PORTC = 0xff ;  // turn off leds
  led = 0xff;

  DDRA = 0xff;
  PORTA = 0xff;
  
  //set up timer 0 for 1 mSec timebase 
  TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
  OCR0A = 249;  		//set the compare re to 250 time ticks
  //set prescalar to divide by 64 
  TCCR0B= 3; //0b00001011;	
  // turn on clear-on-match
  TCCR0A= (1<<WGM01) ;
  
  //initialize ADC
  //init the A to D converter 
   	//channel zero/ left adj /EXTERNAL Aref
   	//!!!CONNECT Aref jumper!!!!
   	//ADMUX = (1<<ADLAR) ;  
	ADMUX = 0;    

   	//enable ADC and set prescaler to 1/128*16MHz=125,000
   	//and clear interupt enable
   	//and start a conversion
   	ADCSRA = (1<<ADEN) | (1<<ADSC) + 7 ; 


  //init the task timer
  time1=t1;
  time2=t2;
  time3=t3;
  time4=t4;
  time5=t5;

  milliCounter = 0;
  secondsCounter = 0;
  minutesCounter = 0;
  hoursCounter = 1;
  ampmFlag = 0;
  /*minL = 0;
  minH = 0;
  hourL = 0;
  hourH = 0;*/
  clock = ~PIND & 0x40;
  digit = 0;


  amin = 0;
  ahour = 12;
  aminL = 0;
  aminH = 0;
  ahourL = 0;
  ahourH = 0;
  a_ampmFlag = 0;

  alarmFlag = 0;
  alarmContinueFlag = 0;
  snoozeFlag = 0;

  //init minutes button push button state machine
  minutesButtonPushState = NoPush;
  minutesButtonFlag = 0;
  hoursButtonPushState = NoPush;
  hoursButtonFlag = 0;
  digitButtonPushState = NoPush;
  digitButtonFlag = 0;
  alarmButtonPushState = NoPush;
  alarmButtonFlag = 0;
  snoozeButtonPushState = NoPush;
  snoozeButtonFlag = 0;

  //init the UART -- uart_init() is in uart.c
  uart_init();
  stdout = stdin = stderr = &uart_str;
  fprintf(stdout,"Starting timers...\n\r");
      
  //crank up the ISRs
  sei();

end  
//==================================================

   
